# mypy: ignore-errors
from abc import ABC, abstractmethod
from typing import Any

import numpy as np
import pandas as pd

from vivarium import Component
from vivarium.framework.engine import Builder


class Force(Component, ABC):
    ##############
    # Properties #
    ##############
    @property
    def configuration_defaults(self) -> dict[str, Any]:
        return {
            self.__class__.__name__.lower(): {
                "max_force": 0.03,
            },
        }

    columns_required = []

    #####################
    # Lifecycle methods #
    #####################

    def setup(self, builder: Builder) -> None:
        self.config = builder.configuration[self.__class__.__name__.lower()]
        self.max_speed = builder.configuration.movement.max_speed

        self.neighbors = builder.value.get_value("neighbors")

        builder.value.register_value_modifier(
            "acceleration",
            modifier=self.apply_force,
            required_resources=self.columns_required + [self.neighbors],
        )

    ##################################
    # Pipeline sources and modifiers #
    ##################################

    def apply_force(self, index: pd.Index, acceleration: pd.DataFrame) -> pd.DataFrame:
        neighbors = self.neighbors(index)
        pop = self.population_view.get(index)
        pairs = self._get_pairs(neighbors, pop)

        raw_force = self.calculate_force(pairs)
        force = self._normalize_and_limit_force(
            force=raw_force,
            velocity=pop[["vx", "vy"]],
            max_force=self.config.max_force,
            max_speed=self.max_speed,
        )

        acceleration.loc[force.index] += force[["x", "y"]]
        return acceleration

    ##################
    # Helper methods #
    ##################

    @abstractmethod
    def calculate_force(self, neighbors: pd.DataFrame):
        pass

    def _get_pairs(self, neighbors: pd.Series, pop: pd.DataFrame):
        pairs = (
            pop.join(neighbors.rename("neighbors"))
            .reset_index()
            .explode("neighbors")
            .merge(
                pop.reset_index(),
                left_on="neighbors",
                right_index=True,
                suffixes=("_self", "_other"),
            )
        )
        pairs["distance_x"] = pairs.x_other - pairs.x_self
        pairs["distance_y"] = pairs.y_other - pairs.y_self
        pairs["distance"] = np.sqrt(pairs.distance_x**2 + pairs.distance_y**2)

        return pairs

    def _normalize_and_limit_force(
        self,
        force: pd.DataFrame,
        velocity: pd.DataFrame,
        max_force: float,
        max_speed: float,
    ):
        normalization_factor = np.where(
            (force.x != 0) | (force.y != 0),
            max_speed / self._magnitude(force),
            1.0,
        )
        force["x"] *= normalization_factor
        force["y"] *= normalization_factor
        force["x"] -= velocity.loc[force.index, "vx"]
        force["y"] -= velocity.loc[force.index, "vy"]
        magnitude = self._magnitude(force)
        limit_scaling_factor = np.where(
            magnitude > max_force,
            max_force / magnitude,
            1.0,
        )
        force["x"] *= limit_scaling_factor
        force["y"] *= limit_scaling_factor
        return force[["x", "y"]]

    def _magnitude(self, df: pd.DataFrame):
        return np.sqrt(np.square(df.x) + np.square(df.y))


class Separation(Force):
    """Push boids apart when they get too close."""

    configuration_defaults = {
        "separation": {
            "distance": 30,
            "max_force": 0.03,
        },
    }

    def calculate_force(self, neighbors: pd.DataFrame):
        # Push boids apart when they get too close
        separation_neighbors = neighbors[neighbors.distance < self.config.distance].copy()
        force_scaling_factor = np.where(
            separation_neighbors.distance > 0,
            ((-1 / separation_neighbors.distance) / separation_neighbors.distance),
            1.0,
        )
        separation_neighbors["force_x"] = (
            separation_neighbors["distance_x"] * force_scaling_factor
        )
        separation_neighbors["force_y"] = (
            separation_neighbors["distance_y"] * force_scaling_factor
        )

        return (
            separation_neighbors.groupby("index_self")[["force_x", "force_y"]]
            .sum()
            .rename(columns=lambda c: c.replace("force_", ""))
        )


class Cohesion(Force):
    """Push boids together."""

    def calculate_force(self, pairs: pd.DataFrame):
        return (
            pairs.groupby("index_self")[["distance_x", "distance_y"]]
            .sum()
            .rename(columns=lambda c: c.replace("distance_", ""))
        )


class Alignment(Force):
    """Push boids toward where others are going."""

    def calculate_force(self, pairs: pd.DataFrame):
        return (
            pairs.groupby("index_self")[["vx_other", "vy_other"]]
            .sum()
            .rename(columns=lambda c: c.replace("v", "").replace("_other", ""))
        )



"""
===================
Resource Management
===================

This module provides a tool to manage dependencies on resources within a
:mod:`vivarium` simulation. These resources take the form of things that can
be created and utilized by components, for example columns in the
:mod:`state table <vivarium.framework.population>`
or :mod:`named value pipelines <vivarium.framework.values>`.

Because these resources need to be created before they can be used, they are
sensitive to ordering. The intent behind this tool is to provide an interface
that allows other managers to register resources with the resource manager
and in turn ask for ordered sequences of these resources according to their
dependencies or raise exceptions if this is not possible.

For more information, see the Resource Management
:ref:`concept note<resource_concept>`.

"""

from vivarium.framework.resource.manager import ResourceInterface, ResourceManager
from vivarium.framework.resource.resource import Resource



from vivarium.framework.lookup.manager import (
    LookupTableInterface,
    LookupTableManager,
    validate_build_table_parameters,
)
from vivarium.framework.lookup.table import LookupTable
from vivarium.types import LookupTableData, ScalarValue



import numpy as np
import pandas as pd
import pytest

from vivarium.framework.event import Event, EventManager


@pytest.fixture
def event_init():
    return {
        "orig": {
            "index": pd.Index(range(10)),
            "user_data": {"key": "value"},
            "time": pd.Timestamp("1/1/2000"),
            "step_size": 12,
        },
        "new_val": {
            "index": pd.Index(range(3)),
            "user_data": {"new_key": "new_value"},
            "time": pd.Timestamp.now(),
            "step_size": 30,
        },
    }


def test_proper_access(event_init):
    # Event attributes are meant to be read-only
    event_data = event_init["orig"]
    e1 = Event(
        event_data["index"],
        event_data["user_data"],
        event_data["time"],
        event_data["step_size"],
    )

    assert (event_data["index"] == e1.index).all()
    assert event_data["user_data"] == e1.user_data
    assert event_data["time"] == e1.time
    assert event_data["step_size"] == e1.step_size

    attribute_data = event_init["new_val"]
    for key, value in attribute_data.items():
        with pytest.raises(AttributeError) as _:
            setattr(e1, key, value)


def test_split_event(event_init):
    event_data = event_init["orig"]
    e1 = Event(
        event_data["index"],
        event_data["user_data"],
        event_data["time"],
        event_data["step_size"],
    )

    new_idx = event_init["new_val"]["index"]
    e2 = e1.split(new_idx)

    assert e1.index is event_data["index"]
    assert e2.index is new_idx


def test_emission(event_init):
    signal = [False]

    def listener(*_, **__):
        signal[0] = True

    manager = EventManager()
    manager.clock = lambda: pd.Timestamp(1990, 1, 1)
    manager.step_size = lambda: pd.Timedelta(30, unit="D")
    manager.add_constraint = lambda f, **kwargs: f
    emitter = manager.get_emitter("test_event")
    manager.register_listener("test_event", listener)
    emitter(event_init["orig"]["index"])

    assert signal[0]

    signal[0] = False

    emitter = manager.get_emitter("test_unheard_event")
    emitter(event_init["new_val"]["index"])
    assert not signal[0]


def test_listener_priority(event_init):
    signal = [False, False, False]

    def listener1(*_, **__):
        signal[0] = True
        assert not signal[1]
        assert not signal[2]

    def listener2(*_, **__):
        signal[1] = True
        assert signal[0]
        assert not signal[2]

    def listener3(*_, **__):
        signal[2] = True
        assert signal[0]
        assert signal[1]

    manager = EventManager()
    manager.clock = lambda: pd.Timestamp(1990, 1, 1)
    manager.step_size = lambda: pd.Timedelta(30, "D")
    manager.add_constraint = lambda f, **kwargs: f
    emitter = manager.get_emitter("test_event")
    manager.register_listener("test_event", listener1, priority=0)
    manager.register_listener("test_event", listener2)
    manager.register_listener("test_event", listener3, priority=9)

    emitter(event_init["orig"]["index"])
    assert np.all(signal)


def test_contains():
    event = "test_event"

    manager = EventManager()
    manager.add_constraint = lambda f, **kwargs: f
    assert event not in manager
    manager.get_emitter(event)
    assert event in manager


def test_list_events():
    manager = EventManager()
    manager.add_constraint = lambda f, **kwargs: f
    _ = manager.get_channel("event1")
    _ = manager.get_emitter("event2")
    _ = manager.register_listener("event3", lambda: None)
    _ = manager.get_listeners("event4")
    assert manager.list_events() == ["event1", "event2", "event3", "event4"]



"""
==================
Randomness Streams
==================

This module provides a wrapper around numpy's randomness system with the intent of coupling
it to vivarium's tools for Common Random Number genereration.


Attributes
----------
RESIDUAL_CHOICE : object
    A probability placeholder to be used in an un-normalized array of weights
    to absorb leftover weight so that the array sums to unity.
    For example::

        [0.2, 0.2, RESIDUAL_CHOICE] => [0.2, 0.2, 0.6]

        
Notes
-----
Currently this object is only used in the `choice` function of this
module.

"""

from __future__ import annotations

import hashlib
from collections.abc import Callable
from typing import TYPE_CHECKING, Any, TypeVar

import numpy as np
import numpy.typing as npt
import pandas as pd
from scipy import stats

from vivarium.framework.randomness.exceptions import RandomnessError
from vivarium.framework.randomness.index_map import IndexMap
from vivarium.framework.resource import Resource
from vivarium.framework.utilities import rate_to_probability
from vivarium.types import ClockTime, NumericArray

if TYPE_CHECKING:
    from vivarium import Component

RESIDUAL_CHOICE = object()

# TODO: Parameterizing pandas objects fails below python 3.12
PandasObject = TypeVar("PandasObject", pd.DataFrame, pd.Series, pd.Index)  # type: ignore [type-arg]


def get_hash(key: str) -> int:
    """Gets a hash of the provided key.

    Parameters
    ----------
    key :
        A string used to create a seed for the random number generator.

    Returns
    -------
        A hash of the provided key.
    """
    max_allowable_numpy_seed = 4294967295  # 2**32 - 1
    return int(hashlib.sha1(key.encode("utf8")).hexdigest(), 16) % max_allowable_numpy_seed


class RandomnessStream(Resource):
    """A stream for producing common random numbers.

    `RandomnessStream` objects provide an interface to Vivarium's
    common random number generation.  They provide a number of methods
    for doing common simulation tasks that require random numbers like
    making decisions among a number of choices.

    Notes
    -----
    Should not be constructed by client code.

    Simulation components get `RandomnessStream` objects by requesting
    them from the builder provided to them during the setup phase.
    I.E.::

        class VivariumComponent:
            def setup(self, builder):
                self.randomness_stream = builder.randomness.get_stream('stream_name')

    """

    def __init__(
        self,
        key: str,
        clock: Callable[[], ClockTime],
        seed: Any,
        index_map: IndexMap,
        # TODO [MIC-5452]: all resources should have a component
        component: Component | None = None,
        initializes_crn_attributes: bool = False,
    ):
        super().__init__("stream", key, component)
        self.key = key
        """The name of the randomness stream."""
        self.clock = clock
        """A way to get the current simulation time."""
        self.seed = seed
        """An extra number used to seed the random number generation."""
        self.index_map = index_map
        """A key-index mapping with a vectorized hash and vectorized lookups."""
        self.initializes_crn_attributes = initializes_crn_attributes
        """A boolean indicating whether the stream is used to initialize CRN attributes."""

    def _key(self, additional_key: Any = None) -> str:
        """Construct a hashable key from this object's state.

        Parameters
        ----------
        additional_key
            Any additional information used to seed random number generation.

        Returns
        -------
            A key to seed random number generation.
        """
        return "_".join([self.key, str(self.clock()), str(additional_key), str(self.seed)])

    def get_draw(self, index: pd.Index[int], additional_key: Any = None) -> pd.Series[float]:
        """Get an indexed set of numbers uniformly drawn from the unit interval.

        Parameters
        ----------
        index
            An index whose length is the number of random draws made
            and which indexes the returned `pandas.Series`.
        additional_key
            Any additional information used to seed random number generation.

        Returns
        -------
            A series of random numbers indexed by the provided `pandas.Index`.

        Notes
        -----
        This is the core of the CRN implementation, allowing for consistent use of random
        numbers across simulations with multiple scenarios.

        See Also
        --------
        https://en.wikipedia.org/wiki/Variance_reduction and
        "Untangling Uncertainty with Common Random Numbers:
        A Simulation Study; A.Flaxman, et. al., Summersim 2017"
        """
        # Return a structured null value if an empty index is passed
        if index.empty:
            return pd.Series(index=index, dtype=float)

        # Initialize a random_state with a seed based on the simulation clock, the
        # decision_point this stream represents, and any additional user-supplied information.
        # This is one pre-condition to reproducibility.
        seed = get_hash(self._key(additional_key))
        random_state = np.random.RandomState(seed=seed)

        # Second, we need to sample a very large chunk of random numbers. The size of the
        # index map is set at the simulation start and is at least 10x the size of the
        # initial population. Which means this is a consistently sampled block of uniformly
        # distributed random numbers, irrespective of the size of the simulation population,
        # which is important if there are scenarios that result in different population sizes
        # through time.
        sample_size = len(self.index_map)
        raw_draws = random_state.random_sample(sample_size)

        if self.initializes_crn_attributes:
            # If we're initializing CRN attributes (i.e. attributes used to identify a
            # simulant across multiple simulations), we can't use the index map yet since
            # these people aren't registered in the index map yet. Instead, we use the draws
            # from our random sample in order. This couples the initialization of CRN
            # attributes to the time step on which they are initialized, meaning interventions
            # that alter the entrance time of a simulant will break the CRN guarantees.
            draws = pd.Series(raw_draws[: len(index)], index=index)
        else:
            # If we're not initializing CRN attributes, we can use the index map to get the
            # correct draws for each simulant. This allows us to use the same CRN attributes
            # across multiple simulations, even if the population size changes.
            draw_index = self.index_map[index]
            draws = pd.Series(raw_draws[draw_index], index=index)

        return draws

    def filter_for_rate(
        self,
        population: PandasObject,
        rate: float | list[float] | tuple[float] | NumericArray | pd.Series[float],
        additional_key: Any = None,
    ) -> PandasObject:
        """Decide an event outcome for each individual from rates.

        Given a population or its index and an array of associated rates for
        some event to happen, we create and return the subpopulation for whom
        the event occurred.

        Parameters
        ----------
        population
            A view on the simulants for which we are determining the
            outcome of an event.
        rate
            A scalar float value or a 1d list of rates of the event under
            consideration occurring which corresponds (i.e.
            `len(population) == len(probability))` to the population view passed
            in. The rates must be scaled to the simulation time-step size either
            manually or as a post-processing step in a rate pipeline. If a
            scalar is provided, it is applied to every row in the population.
        additional_key
            Any additional information used to create the seed.

        Returns
        -------
            The subpopulation of the simulants for whom the event occurred.
            The return type will be the same as type(population).
        """
        return self.filter_for_probability(
            population, rate_to_probability(rate), additional_key
        )

    def filter_for_probability(
        self,
        population: PandasObject,
        probability: float | list[float] | tuple[float] | NumericArray | pd.Series[float],
        additional_key: Any = None,
    ) -> PandasObject:
        """Decide an outcome for each individual from probabilities.

        Given a population or its index and an array of associated probabilities
        for some event to happen, we create and return the subpopulation for
        whom the event occurred.

        Parameters
        ----------
        population
            A view on the simulants for which we are determining the
            outcome of an event.
        probability
            A scalar float value or a 1d list of probabilities of the event
            under consideration occurring which corresponds (i.e.
            `len(population) == len(probability)` to the population view
            passed in. If a scalar is provided, it is applied to every row in
            the population.
        additional_key
            Any additional information used to create the seed.

        Returns
        -------
            The subpopulation of the simulants for whom the event occurred.
            The return type will be the same as type(population).
        """
        if population.empty:
            return population

        if isinstance(population, pd.Index):
            index = population
        else:
            index = population.index

        draws = self.get_draw(index, additional_key)
        mask = np.array(draws < probability)
        return population[mask]

    def choice(
        self,
        index: pd.Index[int],
        choices: list[Any] | tuple[Any] | npt.NDArray[Any] | pd.Series[Any],
        p: (
            list[float | object]
            | tuple[float | object]
            | npt.NDArray[np.number[npt.NBitBase] | np.object_]
            | pd.Series[Any]
            | None
        ) = None,
        additional_key: Any = None,
    ) -> pd.Series[Any]:
        """Decides between a weighted or unweighted set of choices.

        Given a set of choices with or without corresponding weights,
        returns an indexed set of decisions from those choices. This is
        simply a vectorized way to make decisions with some book-keeping.

        Parameters
        ----------
        index
            An index whose length is the number of random draws made
            and which indexes the returned `pandas.Series`.
        choices
            A set of options to choose from.
        p
            The relative weights of the choices.  Can be either a 1-d array of
            the same length as `choices` or a 2-d array with `len(index)` rows
            and `len(choices)` columns.  In the 1-d case, the same set of
            weights are used to decide among the choices for every item in
            the `index`. In the 2-d case, each row in `p` contains a separate
            set of weights for every item in the `index`.
        additional_key
            Any additional information used to seed random number generation.

        Returns
        -------
            An indexed set of decisions from among the available `choices`.

        Raises
        ------
        RandomnessError
            If any row in `p` contains `RESIDUAL_CHOICE` and the remaining
            weights in the row are not normalized or any row of `p contains
            more than one reference to `RESIDUAL_CHOICE`.
        """
        draws = self.get_draw(index, additional_key)
        return _choice(draws, choices, p)

    def sample_from_distribution(
        self,
        index: pd.Index[int],
        distribution: stats.rv_continuous | None = None,
        ppf: Callable[[pd.Series[Any], dict[str, Any]], pd.Series[Any]] | None = None,
        additional_key: Any = None,
        **distribution_kwargs: dict[str, Any],
    ) -> pd.Series[Any]:
        """Given a distribution, returns an indexed set of samples from that
        distribution.

        Parameters
        ----------
        index
            An index whose length is the number of random draws made
            and which indexes the returned `pandas.Series`.
        distribution
            A scipy.stats distribution object.
        ppf
            A function that takes a series of draws and returns a series of samples.
        additional_key
            Any additional information used to seed random number generation.
        distribution_kwargs
            Additional keyword arguments to pass to the distribution's ppf function.

        Returns
        -------
            An indexed set of samples from the provided distribution.
        """
        if ppf is None:
            if distribution is None:
                raise ValueError("Either distribution or ppf must be provided")
            ppf = distribution.ppf
        else:
            if distribution is not None:
                raise ValueError("Only one of distribution or ppf can be provided")

        draws = self.get_draw(index, additional_key)
        samples = pd.Series(ppf(draws, **distribution_kwargs), index=index)  # type: ignore [call-arg]
        return samples

    def __repr__(self) -> str:
        return "RandomnessStream(key={!r}, clock={!r}, seed={!r})".format(
            self.key, self.clock(), self.seed
        )


def _choice(
    draws: pd.Series[float],
    choices: list[Any] | tuple[Any] | npt.NDArray[Any] | pd.Series[Any],
    p: (
        list[float | object]
        | tuple[float | object]
        | npt.NDArray[np.number[npt.NBitBase] | np.object_]
        | pd.Series[Any]
        | None
    ) = None,
) -> pd.Series[Any]:
    """Decides between a weighted or unweighted set of choices.

    Given a set of choices with or without corresponding weights,
    returns an indexed set of decisions from those choices. This is
    simply a vectorized way to make decisions with some book-keeping.

    Parameters
    ----------
    draws
        A uniformly distributed random number for every simulant to make
        a choice for.
    choices
        A set of options to choose from. Choices must be the same for every
        simulant.
    p
        The relative weights of the choices.  Can be either a 1-d array of
        the same length as `choices` or a 2-d array with `len(draws)` rows
        and `len(choices)` columns.  In the 1-d case, the same set of weights
        are used to decide among the choices for every item in the `index`.
        In the 2-d case, each row in `p` contains a separate set of weights
        for every item in the `index`.

    Returns
    -------
        An indexed set of decisions from among the available `choices`.

    Raises
    ------
    RandomnessError
        If any row in `p` contains `RESIDUAL_CHOICE` and the remaining
        weights in the row are not normalized or any row of `p` contains
        more than one reference to `RESIDUAL_CHOICE`.
    """
    # Convert p to normalized probabilities broadcasted over index.
    p_norm = (
        _set_residual_probability(_normalize_shape(p, draws.index))
        if p is not None
        else np.ones((len(draws.index), len(choices)))
    )
    p_norm = p_norm / p_norm.sum(axis=1, keepdims=True)

    p_bins = np.cumsum(p_norm, axis=1)
    # Use the random draw to make a choice for every row in index.
    choice_index = (draws.values[np.newaxis].T > p_bins).sum(axis=1)

    decisions: pd.Series[Any] = pd.Series(np.array(choices)[choice_index], index=draws.index)

    return decisions


def _normalize_shape(
    p: (
        list[float | object]
        | tuple[float | object]
        | npt.NDArray[np.number[npt.NBitBase] | np.object_]
        | pd.Series[Any]
    ),
    index: pd.Index[int] | pd.MultiIndex,
) -> npt.NDArray[np.number[npt.NBitBase] | np.object_]:
    p = np.array(p)
    # We got a 1-d array => same weights for every index.
    if len(p.shape) == 1:
        # Turn 1-d array into 2-d array with same weights in every row.
        p = np.array(np.broadcast_to(p, (len(index), p.shape[0])))
    return p


def _set_residual_probability(
    p: npt.NDArray[np.number[npt.NBitBase] | np.object_],
) -> npt.NDArray[np.float64]:
    """Turns any use of `RESIDUAL_CHOICE` into a residual probability.

    Parameters
    ----------
    p
        Array where each row is a set of probability weights and potentially
        a `RESIDUAL_CHOICE` placeholder.

    Returns
    -------
        Array where each row is a set of normalized probability weights.

    Raises
    ------
    RandomnessError
        If more than one residual choice is supplied for a single set of weights.
    RandomnessError
        If residual choice is supplied with weights that sum to more than 1.
    """
    residual_mask = p == RESIDUAL_CHOICE
    if residual_mask.any():  # I.E. if we have any placeholders.
        if np.any(np.sum(residual_mask, axis=1) - 1):
            msg = (
                "More than one residual choice supplied for a single "
                f"set of weights. Weights: {p}."
            )
            raise RandomnessError(msg)

        p[residual_mask] = 0
        residual_p = 1 - np.sum(p, axis=1)  # Probabilities sum to 1.

        if np.any(residual_p < 0):  # We got un-normalized probability weights.
            msg = (
                "Residual choice supplied with weights that summed to more than 1. "
                f"Weights: {p}."
            )
            raise RandomnessError(msg)

        p[residual_mask] = residual_p
    return p.astype(np.float64)



from typing import Any

import pandas as pd

from vivarium.framework.engine import Builder
from vivarium.framework.results import Observer


class DeathsObserver(Observer):
    """Observes the number of deaths."""

    ##############
    # Properties #
    ##############

    @property
    def columns_required(self) -> list[str] | None:
        return ["alive"]

    #################
    # Setup methods #
    #################

    def register_observations(self, builder: Builder) -> None:
        """We define a newly-dead simulant as one who is 'dead' but who has not
        yet become untracked."""
        builder.results.register_adding_observation(
            name="dead",
            requires_columns=["alive"],
            pop_filter='tracked == True and alive == "dead"',
        )


class YllsObserver(Observer):
    """Observes the years of lives lost."""

    ##############
    # Properties #
    ##############

    @property
    def columns_required(self) -> list[str] | None:
        return ["age", "alive"]

    @property
    def configuration_defaults(self) -> dict[str, Any]:
        return {
            "mortality": {
                "life_expectancy": 80,
            }
        }

    #####################
    # Lifecycle methods #
    #####################

    # noinspection PyAttributeOutsideInit
    def setup(self, builder: Builder) -> None:
        self.life_expectancy = builder.configuration.mortality.life_expectancy

    #################
    # Setup methods #
    #################

    def register_observations(self, builder: Builder) -> None:
        builder.results.register_adding_observation(
            name="ylls",
            requires_columns=["age", "alive"],
            aggregator=self.calculate_ylls,
        )

    def calculate_ylls(self, df: pd.DataFrame) -> float:
        return (self.life_expectancy - df.loc[df["alive"] == "dead", "age"]).sum()



# mypy: ignore-errors
import pandas as pd
from scipy import spatial

from vivarium import Component
from vivarium.framework.engine import Builder
from vivarium.framework.event import Event
from vivarium.framework.population import SimulantData


class Neighbors(Component):
    ##############
    # Properties #
    ##############
    configuration_defaults = {"neighbors": {"radius": 60}}

    columns_required = ["x", "y"]

    #####################
    # Lifecycle methods #
    #####################

    def setup(self, builder: Builder) -> None:
        self.radius = builder.configuration.neighbors.radius

        self.neighbors_calculated = False
        self._neighbors = pd.Series()
        self.neighbors = builder.value.register_value_producer(
            "neighbors", source=self.get_neighbors, required_resources=self.columns_required
        )

    ########################
    # Event-driven methods #
    ########################

    def on_initialize_simulants(self, pop_data: SimulantData) -> None:
        self._neighbors = pd.Series([[]] * len(pop_data.index), index=pop_data.index)

    def on_time_step(self, event: Event) -> None:
        self.neighbors_calculated = False

    ##################################
    # Pipeline sources and modifiers #
    ##################################

    def get_neighbors(self, index: pd.Index) -> pd.Series:
        if not self.neighbors_calculated:
            self._calculate_neighbors()
        return self._neighbors[index]

    ##################
    # Helper methods #
    ##################

    def _calculate_neighbors(self) -> None:
        # Reset our list of neighbors
        pop = self.population_view.get(self._neighbors.index)
        self._neighbors = pd.Series([[] for _ in range(len(pop))], index=pop.index)

        tree = spatial.KDTree(pop[["x", "y"]])

        # Iterate over each pair of simulants that are close together.
        for boid_1, boid_2 in tree.query_pairs(self.radius):
            # .iloc is used because query_pairs uses 0,1,... indexing instead of pandas.index
            self._neighbors.iloc[boid_1].append(self._neighbors.index[boid_2])
            self._neighbors.iloc[boid_2].append(self._neighbors.index[boid_1])



from pathlib import Path

import pytest
import yaml
from click.testing import CliRunner

from tests.framework.results.helpers import HARRY_POTTER_CONFIG
from vivarium.interface.cli import simulate


@pytest.fixture
def runner():
    return CliRunner()


@pytest.fixture
def model_spec(base_config, tmp_path) -> str:
    base_config.update(HARRY_POTTER_CONFIG)
    model_spec = {}
    model_spec["configuration"] = base_config.to_dict()
    model_spec["components"] = {
        "tests.framework.results.helpers": [
            "Hogwarts()",
            "HousePointsObserver()",
            "NoStratificationsQuidditchWinsObserver()",
            "QuidditchWinsObserver()",
            "HogwartsResultsStratifier()",
        ],
    }
    (tmp_path / "model_spec").mkdir()
    filepath = tmp_path / "model_spec" / "model_spec.yaml"
    with open(filepath, "w") as f:
        yaml.dump(model_spec, f)
    return str(filepath)


def test_simulate_run(runner, model_spec, hdf_file_path):

    run_parameters = {param.name for param in simulate.commands["run"].params}
    expected_parameters = {
        "model_specification",
        "artifact_path",
        "results_directory",
        "verbose",
        "quiet",
        "with_debugger",
    }
    different_params = run_parameters.symmetric_difference(expected_parameters)
    if run_parameters.symmetric_difference(expected_parameters):
        raise ValueError(
            f"Missing or unexpected parameters in simulate run: {different_params}"
        )
    output_dir = "/".join(model_spec.split("/")[:-2])
    args = ["run", model_spec, "-o", output_dir, "-i", hdf_file_path]
    sim = runner.invoke(simulate, args)
    assert sim.exit_code == 0
    results_dir = list(Path(output_dir).rglob("*/simulation.log"))[0].parent
    assert {file.name for file in results_dir.rglob("*.parquet")} == {
        "house_points.parquet",
        "quidditch_wins.parquet",
        "no_stratifications_quidditch_wins.parquet",
    }
    with open(results_dir / "metadata.yaml", "r") as f:
        metadata = yaml.safe_load(f)

    assert set(metadata) == set(
        ["input_draw", "random_seed", "simulation_run_time", "artifact_path"]
    )
    assert metadata["simulation_run_time"] > 0.0
    # Ensure '-i' worked
    with open(f"{output_dir}/model_specification.yaml") as f:
        ms = yaml.safe_load(f)
    assert ms["configuration"]["input_data"]["artifact_path"] == str(hdf_file_path)



# mypy: ignore-errors
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation


def plot_boids(simulation, plot_velocity=False):
    width = simulation.configuration.field.width
    height = simulation.configuration.field.height
    pop = simulation.get_population()

    plt.figure(figsize=[12, 12])
    plt.scatter(pop.x, pop.y, color=pop.color)
    if plot_velocity:
        plt.quiver(pop.x, pop.y, pop.vx, pop.vy, color=pop.color, width=0.002)
    plt.xlabel("x")
    plt.ylabel("y")
    plt.axis([0, width, 0, height])
    plt.show()


def plot_boids_animated(simulation):
    width = simulation.configuration.field.width
    height = simulation.configuration.field.height
    pop = simulation.get_population()

    fig = plt.figure(figsize=[12, 12])
    ax = fig.add_subplot(111)
    s = ax.scatter(pop.x, pop.y, color=pop.color)
    plt.xlabel("x")
    plt.ylabel("y")
    plt.axis([0, width, 0, height])

    frames = range(2_000)
    frame_pops = []
    for _ in frames:
        simulation.step()
        frame_pops.append(simulation.get_population()[["x", "y"]])

    def animate(i):
        s.set_offsets(frame_pops[i])

    return FuncAnimation(fig, animate, frames=frames, interval=10)



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from vivarium import Component
    from vivarium.manager import Manager


@dataclass
class Resource:
    """A generic resource representing a node in the dependency graph."""

    resource_type: str
    """The type of the resource."""
    name: str
    """The name of the resource."""
    # TODO [MIC-5452]: all resources should have a component
    component: Component | Manager | None
    """The component that creates the resource."""

    @property
    def resource_id(self) -> str:
        """The long name of the resource, including the type."""
        return f"{self.resource_type}.{self.name}"

    @property
    def is_initialized(self) -> bool:
        """Return True if the resource needs to be initialized."""
        return False


class NullResource(Resource):
    """A node in the dependency graph that does not produce any resources."""

    # TODO [MIC-5452]: all resources should have a component
    def __init__(self, index: int, component: Component | Manager | None):
        super().__init__("null", f"{index}", component)

    @property
    def is_initialized(self) -> bool:
        """Return True if the resource needs to be initialized."""
        return True


class Column(Resource):
    """A resource representing a column in the state table."""

    # TODO [MIC-5452]: all resources should have a component
    def __init__(self, name: str, component: Component | Manager | None):
        super().__init__("column", name, component)

    @property
    def is_initialized(self) -> bool:
        """Return True if the resource needs to be initialized."""
        return True



from __future__ import annotations

import pandas as pd
import pytest
from layered_config_tree.exceptions import ConfigurationError

from tests.helpers import (
    AllColumnsRequirer,
    ColumnCreator,
    ColumnCreatorAndAllRequirer,
    ColumnCreatorAndRequirer,
    ColumnRequirer,
    CustomPriorities,
    DefaultPriorities,
    FilteredPopulationView,
    LookupCreator,
    NoPopulationView,
    OrderedColumnsLookupCreator,
    Parameterized,
    ParameterizedByComponent,
    SingleLookupCreator,
)
from vivarium import Artifact, InteractiveContext
from vivarium.framework.engine import Builder
from vivarium.framework.lookup.table import CategoricalTable, InterpolatedTable, ScalarTable
from vivarium.framework.population import PopulationError


def load_cooling_time(builder: Builder) -> pd.DataFrame:
    return builder.data.load("cooling.time")


def test_unique_component_has_correct_repr():
    component = NoPopulationView()

    # Assert component has the correct repr
    assert component.__repr__() == "NoPopulationView()"


def test_parameterized_component_has_repr_that_incorporates_arguments():
    component = Parameterized("some_value", 5, "another_value")

    # Assert component has the correct repr
    assert component.__repr__() == "Parameterized(p_one=some_value, p_two=5)"


def test_component_by_other_component_has_repr_that_incorporates_arguments():
    arg = Parameterized("some_value", 5, "another_value")
    component = ParameterizedByComponent(arg)

    # Assert component has the correct repr
    expected_repr = "ParameterizedByComponent(param=Parameterized(p_one=some_value, p_two=5))"
    assert component.__repr__() == expected_repr


def test_component_with_no_arguments_has_correct_name():
    component = NoPopulationView()

    # Assert component has the correct name
    assert component.name == "no_population_view"


def test_parameterized_component_has_name_that_incorporates_arguments():
    component = Parameterized("some_value", 5, "another_value")

    # Assert component has the correct name
    assert component.name == "parameterized.some_value.5"


def test_component_by_other_component_has_name_that_incorporates_arguments():
    arg = Parameterized("some_value", 5, "another_value")
    component = ParameterizedByComponent(arg)

    # Assert component has the correct repr
    expected = "parameterized_by_component.'parameterized.some_value.5'"
    assert component.name == expected


def test_component_that_creates_columns_population_view():
    component = ColumnCreator()
    InteractiveContext(components=[component])

    # Assert population view is set and has the correct columns
    assert component.population_view is not None
    assert set(component.population_view.columns) == set(component.columns_created)


def test_component_that_requires_columns_population_view():
    component = ColumnRequirer()
    InteractiveContext(components=[ColumnCreator(), component])

    # Assert population view is set and has the correct columns
    assert component.population_view is not None
    assert set(component.population_view.columns) == set(component.columns_required)


def test_component_that_creates_and_requires_columns_population_view():
    component = ColumnCreatorAndRequirer()
    InteractiveContext(components=[ColumnCreator(), component])

    # Assert population view is set and has the correct columns
    expected_columns = component.columns_required + component.columns_created

    assert component.population_view is not None
    assert set(component.population_view.columns) == set(expected_columns)


@pytest.mark.xfail(reason="This is due to a bug to be fixed by MIC-5373")
def test_component_that_creates_column_and_requires_all_columns_population_view():
    component = ColumnCreatorAndAllRequirer()
    simulation = InteractiveContext(components=[ColumnCreator(), component])
    population = simulation.get_population()

    # Assert population view is set and has the correct columns
    expected_columns = population.columns

    assert component.population_view is not None
    assert set(component.population_view.columns) == set(expected_columns)


def test_component_with_initialization_requirements():
    simulation = InteractiveContext(
        components=[ColumnCreator(), ColumnCreatorAndRequirer()],
    )

    # Assert required resources have been recorded by the ResourceManager
    component_dependencies_list = [
        r.dependencies
        # get all resources in the dependency graph
        for r in simulation._resource.sorted_nodes
        # if the resource is an initializer
        if r.is_initialized
        # its initializer is an instance method
        and hasattr(r.initializer, "__self__")
        # and is a method of ColumnCreatorAndRequirer
        and isinstance(r.initializer.__self__, ColumnCreatorAndRequirer)
    ]
    assert len(component_dependencies_list) == 1
    component_dependencies = component_dependencies_list[0]

    assert "value.pipeline_1" in component_dependencies
    assert "column.test_column_2" in component_dependencies
    assert "stream.stream_1" in component_dependencies


def test_component_that_requires_all_columns_population_view():
    component = AllColumnsRequirer()
    simulation = InteractiveContext(
        components=[ColumnCreator(), ColumnCreatorAndRequirer(), component]
    )
    population = simulation.get_population()

    # Assert population view is set and has the correct columns
    expected_columns = population.columns

    assert component.population_view is not None
    assert set(component.population_view.columns) == set(expected_columns)


def test_component_with_filtered_population_view():
    component = FilteredPopulationView()
    InteractiveContext(components=[ColumnCreator(), component])

    # Assert population view is being filtered using the desired query
    assert component.population_view.query == "test_column_1 == 5 and tracked == True"


def test_component_with_no_population_view():
    component = NoPopulationView()
    InteractiveContext(components=[ColumnCreator(), component])

    # Assert population view is not set
    assert component._population_view is None

    # Assert trying to access the population view raises an error
    with pytest.raises(PopulationError, match=f"'{component.name}' does not have access"):
        _ = component.population_view


def test_component_initializer_is_not_registered_if_not_defined():
    component = NoPopulationView()
    simulation = InteractiveContext(components=[component])

    # Assert that simulant initializer has been registered
    assert (
        component.on_initialize_simulants
        not in simulation._resource.get_population_initializers()
    )


def test_component_initializer_is_registered_and_called_if_defined():
    pop_size = 1000
    component = ColumnCreator()
    expected_pop_view = component.get_initial_state(pd.RangeIndex(pop_size))

    config = {"population": {"population_size": pop_size}}
    simulation = InteractiveContext(components=[component], configuration=config)
    population = simulation.get_population()

    # Assert that simulant initializer has been registered
    assert (
        component.on_initialize_simulants
        in simulation._resource.get_population_initializers()
    )
    # and that created columns are correctly initialized
    pd.testing.assert_frame_equal(population[component.columns_created], expected_pop_view)


def test_listeners_are_not_registered_if_not_defined():
    component = NoPopulationView()
    simulation = InteractiveContext(components=[component])

    post_setup_methods = simulation.get_listeners("post_setup")
    time_step_prepare_methods = simulation.get_listeners("time_step__prepare")
    time_step_methods = simulation.get_listeners("time_step")
    time_step_cleanup_methods = simulation.get_listeners("time_step__cleanup")
    collect_metrics_methods = simulation.get_listeners("collect_metrics")
    simulation_end_methods = simulation.get_listeners("simulation_end")

    for i in range(10):
        assert component.on_post_setup not in set(post_setup_methods.get(i, []))
        assert component.on_time_step_prepare not in set(time_step_prepare_methods.get(i, []))
        assert component.on_time_step not in set(time_step_methods.get(i, []))
        assert component.on_time_step_cleanup not in set(time_step_cleanup_methods.get(i, []))
        assert component.on_collect_metrics not in set(collect_metrics_methods.get(i, []))
        assert component.on_simulation_end not in set(simulation_end_methods.get(i, []))


def test_listeners_are_registered_if_defined():
    component = DefaultPriorities()
    simulation = InteractiveContext(components=[component])

    post_setup_methods = simulation.get_listeners("post_setup")
    time_step_prepare_methods = simulation.get_listeners("time_step__prepare")
    time_step_methods = simulation.get_listeners("time_step")
    time_step_cleanup_methods = simulation.get_listeners("time_step__cleanup")
    collect_metrics_methods = simulation.get_listeners("collect_metrics")
    simulation_end_methods = simulation.get_listeners("simulation_end")

    assert component.on_post_setup in set(post_setup_methods.get(5, []))
    assert component.on_time_step_prepare in set(time_step_prepare_methods.get(5, []))
    assert component.on_time_step in set(time_step_methods.get(5, []))
    assert component.on_time_step_cleanup in set(time_step_cleanup_methods.get(5, []))
    assert component.on_collect_metrics in set(collect_metrics_methods.get(5, []))
    assert component.on_simulation_end in set(simulation_end_methods.get(5, []))


def test_listeners_are_registered_at_custom_priorities():
    component = CustomPriorities()
    simulation = InteractiveContext(components=[component])

    post_setup_methods = simulation.get_listeners("post_setup")
    time_step_prepare_methods = simulation.get_listeners("time_step__prepare")
    time_step_methods = simulation.get_listeners("time_step")
    time_step_cleanup_methods = simulation.get_listeners("time_step__cleanup")
    collect_metrics_methods = simulation.get_listeners("collect_metrics")
    simulation_end_methods = simulation.get_listeners("simulation_end")

    assert component.on_post_setup not in set(post_setup_methods.get(5, []))
    assert component.on_time_step_prepare not in set(time_step_prepare_methods.get(5, []))
    assert component.on_time_step not in set(time_step_methods.get(5, []))
    assert component.on_time_step_cleanup not in set(time_step_cleanup_methods.get(5, []))
    assert component.on_collect_metrics not in set(collect_metrics_methods.get(5, []))
    assert component.on_simulation_end not in set(simulation_end_methods.get(5, []))

    assert component.on_post_setup in set(post_setup_methods.get(8, []))
    assert component.on_time_step_prepare in set(time_step_prepare_methods.get(7, []))
    assert component.on_time_step in set(time_step_methods.get(2, []))
    assert component.on_time_step_cleanup in set(time_step_cleanup_methods.get(3, []))
    assert component.on_collect_metrics in set(collect_metrics_methods.get(6, []))
    assert component.on_simulation_end in set(simulation_end_methods.get(1, []))


def test_component_configuration_gets_set():
    without_config = ColumnCreator()
    with_config = ColumnRequirer()

    column_requirer_config = {
        "column_requirer": {"test_configuration": "some_config_value"},
    }

    sim = InteractiveContext(components=[with_config, without_config], setup=False)
    sim.configuration.update(column_requirer_config)

    assert without_config.configuration is None
    assert with_config.configuration is None

    sim.setup()

    assert without_config.configuration is None
    assert with_config.configuration is not None
    assert with_config.configuration.to_dict() == column_requirer_config["column_requirer"]


def test_component_lookup_table_configuration(hdf_file_path):
    # Tests that lookup tables are created correctly based on their configuration

    favorite_team = pd.DataFrame(
        {"value": ["team_1", "team_2", "team_3"], "test_column_1": [1, 2, 3]}
    ).set_index("test_column_1")
    favorite_number = pd.DataFrame(
        {
            "value": ["number_1", "number_2", "number_3"],
            "test_column_3_start": [0, 1, 2],
            "test_column_3_end": [1, 2, 3],
        }
    ).set_index(["test_column_3_start", "test_column_3_end"])
    favorite_color = pd.DataFrame(
        {
            "value": ["color_1", "color_2", "color_3"],
            "test_column_2": ["value_1", "value_2", "value_3"],
            "test_column_3_start": [0, 1, 2],
            "test_column_3_end": [1, 2, 3],
        }
    ).set_index(["test_column_2", "test_column_3_start", "test_column_3_end"])
    cooling_time = pd.DataFrame(
        {"value": [0.1, 0.9, 0.2], "test_column_1": [1, 2, 3]}
    ).set_index("test_column_1")
    artifact_data = {
        "simulants.favorite_team": favorite_team,
        "simulants.favorite_color": favorite_color,
        "simulants.favorite_number": favorite_number,
        "cooling.time": cooling_time,
    }
    artifact = Artifact(hdf_file_path)
    for key, data in artifact_data.items():
        artifact.write(key, data)

    component = LookupCreator()
    sim = InteractiveContext(components=[component], setup=False)
    sim.configuration.update(
        {
            "input_data": {"artifact_path": hdf_file_path},
        },
    )
    sim.setup()

    # Assertions for specific lookup tables
    expected_tables = {
        "favorite_team",
        "favorite_color",
        "favorite_number",
        "favorite_scalar",
        "favorite_list",
        "baking_time",
        "cooling_time",
    }
    assert expected_tables == set(component.lookup_tables.keys())

    # Check for correct columns in lookup tables
    assert component.lookup_tables["favorite_team"].key_columns == ["test_column_1"]
    assert not component.lookup_tables["favorite_team"].parameter_columns
    assert component.lookup_tables["favorite_color"].key_columns == ["test_column_2"]
    assert component.lookup_tables["favorite_color"].parameter_columns == ["test_column_3"]
    assert isinstance(component.lookup_tables["favorite_scalar"], ScalarTable)
    assert isinstance(component.lookup_tables["favorite_list"], ScalarTable)
    assert component.lookup_tables["favorite_list"].value_columns == ["column_1", "column_2"]
    assert isinstance(component.lookup_tables["baking_time"], ScalarTable)
    assert component.lookup_tables["cooling_time"].key_columns == ["test_column_1"]
    assert not component.lookup_tables["cooling_time"].parameter_columns

    # Check for correct data in lookup tables
    assert component.lookup_tables["favorite_team"].data.equals(favorite_team.reset_index())
    assert component.lookup_tables["favorite_color"].data.equals(favorite_color.reset_index())
    assert component.lookup_tables["favorite_scalar"].data == 0.4
    assert component.lookup_tables["favorite_list"].data == [9, 4]
    assert component.lookup_tables["baking_time"].data == 0.5
    assert component.lookup_tables["cooling_time"].data.equals(cooling_time.reset_index())


@pytest.mark.parametrize(
    "configuration, match, error_type",
    [
        (
            {"favorite_color": "key.not.in.artifact"},
            "Error building lookup table 'favorite_color'. "
            "Failed to find key 'key.not.in.artifact' in artifact.",
            ConfigurationError,
        ),
        (
            {"favorite_color": "not.a.real.module::load_color"},
            "Error building lookup table 'favorite_color'. "
            "Unable to find module 'not.a.real.module'",
            ConfigurationError,
        ),
        (
            {"favorite_color": "self::non_existent_loader_function"},
            "Error building lookup table 'favorite_color'. There is no method "
            "'non_existent_loader_function' for the component "
            "single_lookup_creator.",
            ConfigurationError,
        ),
        (
            {"favorite_color": "vivarium::non_existent_loader_function"},
            "Error building lookup table 'favorite_color'. There is no method "
            "'non_existent_loader_function' for the module 'vivarium'.",
            ConfigurationError,
        ),
    ],
)
def test_failing_component_lookup_table_configurations(
    configuration, match, error_type, hdf_file_path
):
    component = SingleLookupCreator()
    sim = InteractiveContext(components=[component], setup=False)
    override_config = {
        "input_data": {"artifact_path": hdf_file_path},
        component.name: {"data_sources": configuration},
    }
    sim.configuration.update(override_config)
    with pytest.raises(error_type, match=match):
        sim.setup()


@pytest.mark.parametrize(
    "table", ["ordered_columns_categorical", "ordered_columns_interpolated"]
)
def test_value_column_order_is_maintained(table):
    """Tests that the order of value columns is maintained when creating a LookupTable.

    Notes
    -----
    This test is a bit of a hack. We found an issue where the order of value columns
    was changing due to casting the value columns as a set on the back end (which
    does not guarantee order). The problem is that we can't actually guarantee
    that casting as a set will change the order either. However, with a large
    enough number of value columns, it seems likely that the order will change.
    """
    component = OrderedColumnsLookupCreator()
    sim = InteractiveContext(components=[component])
    lookup_table = component.lookup_tables[table]
    assert isinstance(
        lookup_table, CategoricalTable if "categorical" in table else InterpolatedTable
    )
    data = lookup_table(sim.get_population().index)
    assert list(data.columns) == ["one", "two", "three", "four", "five", "six", "seven"]



"""
===================
Randomness IndexMap
===================

The :class:`IndexMap` is an internal abstraction used by the randomness system to help align
random numbers for the same simulants across multiple simulations. It's key idea is to take
a set of static identifying characteristics about a simulant and hash them to a consistent
positional index within a stream of seeded random numbers.

"""

from __future__ import annotations

from datetime import datetime
from typing import Any

import numpy as np
import pandas as pd
import pandas.api.types as pdt

from vivarium.framework.randomness.exceptions import RandomnessError
from vivarium.types import ClockTime


class IndexMap:
    """A key-index mapping with a vectorized hash and vectorized lookups."""

    SIM_INDEX_COLUMN = "simulant_index"
    TEN_DIGIT_MODULUS = 10_000_000_000

    def __init__(self, key_columns: list[str] | None = None, size: int = 1_000_000):
        self._use_crn = bool(key_columns)
        self._key_columns = key_columns if key_columns else []
        self._map: pd.Series[int] | None = None
        """The mapping between the key columns and the randomness index."""
        self._size = size

    def update(self, new_keys: pd.DataFrame, clock_time: ClockTime) -> None:
        """Adds the new keys to the mapping.

        Parameters
        ----------
        new_keys
            A pandas DataFrame indexed by the simulant index and columns corresponding to
            the randomness system key columns.
        clock_time
            The simulation clock time. Used as the salt during hashing to
            minimize inter-simulation collisions.
        """
        if new_keys.empty or not self._use_crn:
            return  # Nothing to do

        new_mapping_index, final_mapping_index = self._parse_new_keys(new_keys)

        final_keys = final_mapping_index.droplevel(self.SIM_INDEX_COLUMN)
        if len(final_keys) != len(final_keys.unique()):
            raise RandomnessError("Non-unique keys in index")

        final_mapping = self._build_final_mapping(new_mapping_index, clock_time)

        # Tack on the simulant index to the front of the map.
        final_mapping.index = final_mapping.index.join(final_mapping_index).reorder_levels(
            [self.SIM_INDEX_COLUMN] + self._key_columns
        )
        final_mapping = final_mapping.sort_index(level=self.SIM_INDEX_COLUMN)
        self._map = final_mapping

    def _parse_new_keys(self, new_keys: pd.DataFrame) -> tuple[pd.Index[Any], pd.Index[Any]]:
        """Parses raw new keys into the mapping index.

        This returns a tuple of the new and final mapping indices. Both are pandas
        indices with a level for the index assigned by the population system and
        additional levels for the key columns associated with the simulant index. The
        new mapping index contains only the values for the new keys and the final mapping
        combines the existing mapping and the new mapping index.

        Parameters
        ----------
        new_keys
            A pandas DataFrame indexed by the simulant index and columns corresponding to
            the randomness system key columns.

        Returns
        -------
            A tuple of the new mapping index and the final mapping index.
        """
        keys = new_keys.copy()
        keys.index.name = self.SIM_INDEX_COLUMN
        new_mapping_index = keys.set_index(self._key_columns, append=True).index

        if self._map is None:
            final_mapping_index = new_mapping_index
        else:
            final_mapping_index = self._map.index.append(new_mapping_index)  # type: ignore [no-untyped-call]
        return new_mapping_index, final_mapping_index

    def _build_final_mapping(
        self, new_mapping_index: pd.Index[Any], clock_time: ClockTime
    ) -> pd.Series[int]:
        """Builds a new mapping between key columns and the randomness index from the
        new mapping index and the existing map.

        Parameters
        ----------
        new_mapping_index
            An index with a level for the index assigned by the population system and
            additional levels for the key columns associated with the simulant index.
        clock_time
            The simulation clock time. Used as the salt during hashing to
            minimize inter-simulation collisions.

        Returns
        -------
            The new mapping incorporating the updates from the new mapping index and
            resolving collisions.
        """
        new_key_index = new_mapping_index.droplevel(self.SIM_INDEX_COLUMN)
        mapping_update = self._hash(new_key_index, salt=clock_time)
        if self._map is None:
            current_map = mapping_update
        else:
            old_map = self._map.droplevel(self.SIM_INDEX_COLUMN)
            current_map = pd.concat([old_map, mapping_update])

        return self._resolve_collisions(new_key_index, current_map)

    def _resolve_collisions(
        self,
        new_key_index: pd.Index[Any],
        current_mapping: pd.Series[int],
    ) -> pd.Series[int]:
        """Resolves collisions in the new mapping by perturbing the hash.

        Parameters
        ----------
        new_key_index
            The index of new key attributes to hash.
        current_mapping
            The new mapping incorporating the updates from the new mapping index with
            collisions unresolved.

        Returns
        -------
            The new mapping incorporating the updates from the new mapping index and
            resolving collisions.
        """
        current_mapping = current_mapping.drop_duplicates()
        collisions = new_key_index.difference(current_mapping.index)
        salt = 1
        while not collisions.empty:
            mapping_update = self._hash(collisions, salt)
            current_mapping = pd.concat([current_mapping, mapping_update]).drop_duplicates()
            collisions = mapping_update.index.difference(current_mapping.index)
            salt += 1
        return current_mapping

    def _hash(self, keys: pd.Index[Any], salt: ClockTime = 0) -> pd.Series[int]:
        """Hashes the index into an integer index in the range [0, self.stride]

        Parameters
        ----------
        keys
            The new index to hash.
        salt
            Value used to perturb the hash in a deterministic way. Useful
            in dealing with collisions.

        Returns
        -------
            A pandas series indexed by the given keys and whose values take on
            integers in the range [0, len(self)].  Duplicates may appear and
            should be dealt with by the calling code.
        """
        key_frame = keys.to_frame()
        new_map = pd.Series(0, index=keys)
        salt_series = self._convert_to_ten_digit_int(pd.Series(salt, index=keys))

        for _i, column_name in enumerate(key_frame.columns):
            column = self._convert_to_ten_digit_int(key_frame[column_name])

            primes = [2, 3, 5, 7, 11, 13, 17, 19, 23, 27]
            out = pd.Series(1, index=column.index)
            for idx, p in enumerate(primes):
                # numpy will almost always overflow here, but it is equivalent
                # to modding out by 2**64.  Since it's much much larger than
                # our map size the amount of additional periodicity this
                # introduces is pretty trivial.
                out *= np.power(p, self._digit(column, idx))
            new_map += out + salt_series

        return new_map % len(self)

    def _convert_to_ten_digit_int(
        self, column: pd.Series[datetime | int | float]
    ) -> pd.Series[int]:
        """Converts a column of datetimes, integers, or floats into a column
        of 10 digit integers.

        Parameters
        ----------
        column
            A series of datetimes, integers, or floats.

        Returns
        -------
            A series of ten digit integers based on the input data.

        Raises
        ------
        RandomnessError
            If the column contains data that is neither a datetime-like nor
            numeric.
        """
        if pdt.is_datetime64_any_dtype(column):
            integers = self._clip_to_seconds(column.astype(np.int64))
        elif pdt.is_integer_dtype(column):
            if not len(column >= 0) == len(column):
                raise RandomnessError(
                    "Values in integer columns must be greater than or equal to zero."
                )
            integers = self._spread(column.astype(int))
        elif pdt.is_float_dtype(column):
            integers = self._shift(column.astype(float))
        else:
            raise RandomnessError(
                f"Unhashable column type {type(column.iloc[0])}. "
                "IndexMap accepts datetime like columns and numeric columns."
            )
        return integers

    @staticmethod
    def _digit(m: pd.Series[int], n: int) -> pd.Series[int]:
        """Returns the nth digit of each number in m."""
        nth_digits: pd.Series[int] = (m // (10**n)) % 10
        return nth_digits

    @staticmethod
    def _clip_to_seconds(m: pd.Series[int]) -> pd.Series[int]:
        """Clips UTC datetime in nanoseconds to seconds."""
        return m // pd.Timedelta(1, unit="s").value

    def _spread(self, m: pd.Series[int]) -> pd.Series[int]:
        """Spreads out integer values to give smaller values more weight."""
        return (m * 111_111) % self.TEN_DIGIT_MODULUS

    def _shift(self, m: pd.Series[float]) -> pd.Series[int]:
        """Shifts floats so that the first 10 decimal digits are significant."""
        out = m % 1 * self.TEN_DIGIT_MODULUS // 1
        return out.astype("int64")

    def __getitem__(self, index: pd.Index[int]) -> np.ndarray[int, Any]:
        if self._use_crn:
            if self._map is None:
                raise RandomnessError("IndexMap is empty")
            else:
                return self._map.loc[index].to_numpy()
        else:
            return index.values

    def __len__(self) -> int:
        return self._size

    def __repr__(self) -> str:
        return "IndexMap({})".format("\n         ".join(repr(self._map).split("\n")))
